unit UDMod;

interface

uses
  SysUtils, Classes, UDevil, StdCtrls, ExtCtrls, Controls, ComCtrls, UJDRObjects,
  ScktComp, Dialogs, forms;

const
  // Les diffrents icones de l'imagemap
  CICON_MAP     = 0;
  CICON_EVENT   = 1;
  CICON_AERA    = 2;
  CICON_PNJ     = 3;
  CICON_PLAYER  = 4;
  CICON_OBJECTS = 5;
  CICON_GREEN   = 6;
  CICON_YELLOW  = 7;
  CICON_RED     = 8;
  CICON_TRASH   = 9;
  CICON_MAGNIF  = 10;
  CICON_RENAME  = 11;
  CICON_FOLDER  = 12;
  CICON_LOCK    = 13;

  // Index des items de controle
  CCTRL_AFF     = 0;
  CCTRL_CLI     = 1;
  CCTRL_SER     = 2;

  // Tag des images
  CIMGTAG_MAP   = 0;
  CIMGTAG_PNJ   = 1;

  // Temps de lock en 1/4s
  CLOCK_DEFAULT = 2500; // (environs 10 minutes avant timeout)
  CLOCK_MOVE    = 8;

  // LineFeed
  LF = char(10);


type
  TDMod = class(TDataModule)
    Client: TClientSocket;
    Server: TServerSocket;
    SaveDialog: TSaveDialog;
    OpenDialog: TOpenDialog;
    Timer: TTimer;
    procedure DataModuleCreate(Sender: TObject);
    procedure DataModuleDestroy(Sender: TObject);
    procedure ServerAccept(Sender: TObject; Socket: TCustomWinSocket);
    procedure ServerClientRead(Sender: TObject; Socket: TCustomWinSocket);
    procedure ClientRead(Sender: TObject; Socket: TCustomWinSocket);
    procedure ClientDisconnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure ServerClientConnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ServerClientDisconnect(Sender: TObject;
      Socket: TCustomWinSocket);
    procedure ServerListen(Sender: TObject; Socket: TCustomWinSocket);
    procedure ClientConnect(Sender: TObject; Socket: TCustomWinSocket);
    procedure ClientError(Sender: TObject; Socket: TCustomWinSocket;
      ErrorEvent: TErrorEvent; var ErrorCode: Integer);
    procedure TimerTimer(Sender: TObject);
  private
    { Private declarations }
    TempBuffer: string;
    Buffer: TStringList;
    NbConnected: Integer;
    ServerFile: Textfile;

  public
    { Public declarations }
    Devil: TDevilDLL;
    AllObjects: TTreeNodes;
    TimeCnt: Integer;

    procedure LoadDirIntoCombo(Dir: string; Combo: TComboBox);
    function CStr(Value: Integer): String;
    function RStr(Value: real): String;
    function CVal(Value: String): real;
    function Parser(Str: string; Index: integer; Separator: char = ';'): string;
    function ParseComment(Str: string): Tstrings;
    function FindJDRObject(VObjID: Integer): TJDRObject;
    function BadChars(Str: string): string;

    procedure LoadDatas(Target: TTreeNode; Datas: TStringList);
    function MoveMovable(Serial: String): TMovable;
    function ModifyObj(Serial: String): TJDRObject;
    procedure LockObj(JDRObject: TJDRObject);
    procedure UnLockObj(JDRObject: TJDRObject);

    // Serveur
    procedure OpenServer(Port: Integer);
    procedure CloseServer;
    procedure OpenClient(IP: String; Port: Integer);
    procedure CloseClient;
    procedure ToServer(Str: String);

    procedure Hash(var ToHash: String; StringList: TStringList; Separator: Char);
    procedure DoBufferCmd(VBuffer: TStringList);
  end;

var
  DMod: TDMod;

implementation

uses StrUtils, Main, Math;

{$R *.dfm}

procedure TDMod.DataModuleCreate(Sender: TObject);
begin
  // Initialisation de Devil
  Devil := TDevilDLL.Create( 'DevIL.dll', 'ILU.dll', 'ILUT.dll');
  TempBuffer := '';
  Buffer := TStringList.Create;
end;

procedure TDMod.DataModuleDestroy(Sender: TObject);
begin
  // Destruction de Devil
  Devil.Destroy;
  Buffer.Destroy;
end;



// ------ MISC -------

// Charge une liste de fichiers dans une ComboBox
procedure TDMod.LoadDirIntoCombo(Dir: string; Combo: TComboBox);
var Code: Integer;
    Files: TSearchRec;
begin
  Code := FindFirst( Dir, faAnyFile, Files);
  while ( Code = 0 ) do begin
    Combo.AddItem( Files.Name, nil);
    Code := FindNext( Files );
  end;
  FindClose( Files );
end;

function TDMod.CStr(Value: integer): String;
begin
  Str(Value, Result);
end;

function TDMod.RStr(Value: real): String;
begin
  Str(Value, Result);
end;

function TDmod.CVal(Value: String): real;
var Code : integer;
begin
  Val(Value,Result,Code);
end;

function TDMod.Parser(Str: string; Index: integer; Separator: char = ';'): string;
var Offset: integer;
    LenStr: integer;
    VirCount: integer;
begin
  VirCount := 0;
  Offset := 1;
  while ((VirCount < Index) and (Offset < length(Str))) do begin
    if (Str[Offset] = Separator) then  inc(VirCount);
    inc(Offset);
  end;
  LenStr := 0;
  while (Str[Offset+LenStr] <> Separator) and (Offset+LenStr < length(Str)) do begin
    inc(LenStr);
  end;
  Result := MidStr(Str, Offset, LenStr);
end;

function TDMod.ParseComment(Str: string): Tstrings;
var StrList: TStringList;
    Offset: integer;
    Compte: integer;
begin
  StrList := TStringList.Create;
  Compte :=0;
  for Offset := 1 to length(Str) do
    if Str[Offset]='|' then inc(Compte);
  for Offset :=0 to Compte-1 do
    strlist.Append(Parser(Str,Offset,'|'));
  result:=StrList;
end;

// Trouve un ID objet
function TDMod.FindJDRObject(VObjID: Integer): TJDRObject;
var Boucle: Integer;
    Data: Pointer;
    JDRObject: TJDRObject;
begin
  Boucle := 0;
  JDRObject := nil;
  While ((not assigned(JDRObject)) and (Boucle < AllObjects.count)) do begin
    Data := AllObjects[Boucle].Data;
    if (assigned(Data)) then
      if (TJDRObject(Data).ObjID = VObjID) then JDRObject := Data;
    inc(Boucle);
  end;
  Result := JDRObject;
end;

function TDmod.BadChars(Str: string): string;
var Boucle: Integer;
begin
  for Boucle:=1 to length(Str) do begin
    if (Str[Boucle]=';') then Str[Boucle]:=','
    else if (Str[Boucle]='|') then Str[Boucle]:='/';
  end;
  Result:=Str;
end;

procedure TDMod.LockObj(JDRObject: TJDRObject);
begin
  if (not (JDRObject is TDirectory)) then
    ToServer('!LOCK' + LF + CStr(JDRObject.ObjID) + LF + '!END-LOCK' + LF);
end;
procedure TDMod.UnLockObj(JDRObject: TJDRObject);
begin
  if (not (JDRObject is TDirectory)) then
    ToServer('!UNLOCK' + LF + CStr(JDRObject.ObjID) + LF + '!END-UNLOCK' + LF);
end;

// -------------------------
// -- Connexions Internet --
// -------------------------

function TDMod.MoveMovable(Serial: String): TMovable;
var Movable: TMovable;
begin
  Result := nil;
  Movable := TMovable(FindJDRObject(round(CVal(Parser(Serial,2)))));
  if (assigned(Movable)) then begin
    Movable.MapID := round(CVal(Parser(Serial,5)));
    if (assigned(Movable.Map)) then begin
      Movable.PercX := CVal(Parser(Serial,6));
      Movable.PercY := CVal(Parser(Serial,7));
    end;
    Result:= Movable;
  end;
end;

// Buffers

procedure TDmod.Hash(var ToHash: String; StringList: TStringList; Separator: Char);
var Offset: Integer;
begin
  repeat
    Offset := Pos(Separator, ToHash);
    if (Offset > 0) then begin
      StringList.Append(LeftStr(ToHash, Offset-1));
      ToHash := RightStr(ToHash, length(ToHash)-Offset);
    end;
  until (Offset = 0);
end;

function TDMod.ModifyObj(Serial: String): TJDRObject;
var JDRObject: TJDRObject;
    Datas: TStringList;
begin
  Result := nil;
  if ( Parser(Serial,0) = 'Obj' ) then begin
    JDRObject := TJDRObject(FindJDRObject(round(CVal(Parser(Serial,2)))));
    if (assigned(JDRObject)) then begin
      Datas := TStringList.Create;
      Hash(Serial, Datas, ';');
      if (JDRObject.LoadSerial(Datas)) then Result := JDRObject;
      Datas.Free;
    end;
  end;
end;

procedure TDmod.DoBufferCmd(VBuffer: TStringList);
var EndIndex, Boucle: Integer;
    Datas: TStringList;
    Movable: TMovable;
    JDRObject: TJDRObject;
    Directory: TDirectory;
    Ok: Boolean;
    ExpandSave: Boolean;
    Target: TTreeNode;

  function FindEnd(Cmd: String): Integer;
  var Boucle: Integer;
      Count: Integer;
  begin
    Result:=0;
    Boucle:=0;
    Count:=VBuffer.Count;
    While ((Boucle < Count) and (Result=0)) do begin
      if (VBuffer[Boucle]=Cmd) then Result := Boucle;
      inc(Boucle);
    end;
  end;
  procedure DelBuf(Count: Integer);
  begin
    while (Count > 0) do begin
      VBuffer.Delete(0);
      dec(Count);
    end;
  end;

begin
  Ok := True;

  // Normalement inutile, scurit anti-freeze
  while (Ok and (VBuffer.Count > 0)) do begin
    Ok := False;
//    while ((VBuffer.Count > 0) and (LeftStr(VBuffer[0],1) <> '!')) do VBuffer.Delete(0);

    if (VBuffer.Count > 0) then begin
      // Chargement des donnes
      if (VBuffer[0] = '!LOAD') then begin
        EndIndex := FindEnd('!END-LOAD');
        if (EndIndex > 0) then begin
          Ok := True;
          Datas := TStringList.Create;
          for Boucle:=0 to EndIndex-1 do begin
            DelBuf(1);
            Datas.Append(VBuffer[0]);
          end;
          DelBuf(1);
          LoadDatas(nil, Datas);
          Datas.Destroy;
        end;
      end else if ((VBuffer[0] = '!MODIFY') and (FindEnd('!END-MODIFY') = 2)) then begin
        Ok := True;
        JDRObject := ModifyObj(VBuffer[1]);
        DelBuf(3);

        if (JDRObject is TMovable) then begin
          Movable := TMovable(JDRObject);
          Movable.Locked := CLOCK_MOVE;
          Movable.SetIcon;
          if (FormMain.ActiveMap = Movable.Map) then
            Movable.Display(FormMain.ImageMap)
          else
            Movable.Hide;
        end else if (JDRObject is TEvent) then FormMain.DrawTimeMap;
      end else if ((VBuffer[0] = '!LOCK') and (FindEnd('!END-LOCK') = 2)) then begin
        Ok := True;
        JDRObject := FindJDRObject(Round(CVal(VBuffer[1])));
        DelBuf(3);
        JDRObject.Locked := CLOCK_DEFAULT;
        JDRObject.SetIcon;
      end else if ((VBuffer[0] = '!UNLOCK') and (FindEnd('!END-UNLOCK') = 2)) then begin
        Ok := True;
        JDRObject := FindJDRObject(Round(CVal(VBuffer[1])));
        DelBuf(3);
        JDRObject.Locked := 0;
        JDRObject.SetIcon;
      end else if ((VBuffer[0] = '!CHAT') and (FindEnd('!END-CHAT') = 2)) then begin
        Ok := True;
        FormMain.ChatMemo.Lines.Append(VBuffer[1]);
        DelBuf(3);
      end else if ((VBuffer[0] = '!CONNECT') and (FindEnd('!END-CONNECT') = 2)) then begin
        Ok := True;
        FormMain.ChatMemo.Lines.Append(VBuffer[1] + ' connected...');
        DelBuf(3);
      end else if ((VBuffer[0] = '!MOVE') and (FindEnd('!END-MOVE') = 4)) then begin
        Ok := True;
        Target := FindJDRObject(Round(CVal(VBuffer[3]))).SelfNode;
        if (VBuffer[1] = 'CHILD') then begin
          ExpandSave := Target.Expanded;
          FindJDRObject(Round(CVal(VBuffer[2]))).SelfNode.MoveTo(Target, naAddChild)
        end else begin
          ExpandSave := Target.Parent.Expanded;
          FindJDRObject(Round(CVal(VBuffer[2]))).SelfNode.MoveTo(Target, naAdd);
        end;
        Target.Expanded := ExpandSave;
        AllObjects.AlphaSort(True);
        DelBuf(5);
      end else if ((VBuffer[0] = '!CREATE') and (FindEnd('!END-CREATE') = 3)) then begin
        Ok := True;
        Datas := TStringList.Create;
        Datas.Append(VBuffer[2]);
        LoadDatas(FindJDRObject(Round(CVal(VBuffer[1]))).SelfNode, Datas);
        Datas.Free;
        AllObjects.AlphaSort(True);
        DelBuf(4);
      end else if ((VBuffer[0] = '!NEWDIR') and (FindEnd('!END-NEWDIR') = 2)) then begin
        Ok := True;
        Directory := FormMain.Newdirectory;
        Directory.ObjID := Round(CVal(Parser(VBuffer[1], 2)));
        Directory.Name := Parser(VBuffer[1], 3);
        Directory.SelfNode.Text := Directory.Name;
        AllObjects.AlphaSort(True);
        DelBuf(3);
      end else if ((VBuffer[0] = '!DELETE') and (FindEnd('!END-DELETE') = 2)) then begin
        Ok := True;
        JDRObject := FindJDRObject(Round(CVal(VBuffer[1])));
        if (assigned(JDRObject)) then begin
          if (JDRObject = FormMain.ActiveMap) then FormMain.FreeMap;
          JDRObject.Free;
        end;
        DelBuf(3);
      end else if ((VBuffer[0] = '!CLOCK') and (FindEnd('!END-CLOCK') = 2)) then begin
        Ok := True;
        FormMain.LoadClocksSerial(VBuffer[1]);
        DelBuf(3);
      end else if (VBuffer[0] = '!TIMEDOWN') then begin
        Ok := True;
        FormMain.TimePlayBtn.Down := True;
        FormMain.TimePanelColor;
        DelBuf(1);
      end else if (VBuffer[0] = '!TIMEUP') then begin
        Ok := True;
        FormMain.TimePlayBtn.Down := False;
        FormMain.TimePanelColor;
        DelBuf(1);
        FormMain.DrawTimeMap;
      end;
    end;
  end;
end;

// Load / a partir d'une TStringList

procedure TDmod.LoadDatas(Target: TTreeNode; Datas: TStringList);
var Boucle: Integer;
    TypeObj, Str: string;
    Directory: TDirectory;
    Map: TMap;
    Pnj: TPnj;
    Player: TPlayer;
    Aera: TAera;
    Event: TEvent;
    TempNode, Node: TTreeNode;
begin
  // Libration des objets
  TempNode := nil;
  if (not assigned(Target)) then begin
    FormMain.FreeMap;
    FormMain.FreeObjects;
    TempNode := FormMain.GameObj.Items.Add(nil, '** Loading **');
    Node := TempNode;
  end else Node := Target;

  // Premire passe: cration des objets
  for Boucle:=0 to Datas.count-1 do begin
    // lire une ligne
    Str := Datas[Boucle];
    // la decoder
    if ( Parser(Str,0) = 'Clk' ) then begin
      FormMain.LoadClocksSerial(Str);
    end else if ( Parser(Str,0) = 'Obj' ) then begin

      TypeObj := Parser(Str,1);
      if ( TypeObj = 'TDirectory' ) then begin
        Directory := TDirectory.Create(Node, Parser(Str,3));
        Directory.ObjID := round(CVal(Parser(Str,2)));
        Node := Directory.SelfNode;

      end else if ( TypeObj = 'TMap' ) then begin
      // MAP
      // Obj;TMap;3;NewMap;Comments|;Aimee Major - 2000 - Shirt.jpg;15;
        Map := TMap.Create(Node, Parser(Str,3), ParseComment(Parser(Str,4)), Parser(Str,5), Round(DMod.CVal(Parser(Str,6))));
        Map.ObjID := round(CVal(Parser(Str,2)));

      end else if ( TypeObj = 'TAera' ) then begin
      // AERA
      // Obj;TAera;4;NewAera;Comments|;;0;0;
        Aera := TAera.Create(Node, Parser(Str,3), ParseComment(Parser(Str,4)), FormMain.AeraThumb, FormMain, FormMain.PanelMap, FormMain.ImageMapDragOver, FormMain.ImageThumbMouseMove);
        Aera.ObjID := round(CVal(Parser(Str,2)));

      end else if ( TypeObj = 'TPnj' ) then begin
      // PNJ
      // Obj;TPnj;5;NewPnj;Comments|;;0;0;N;5
        Pnj := TPnj.Create(Node, Parser(Str,3), ParseComment(Parser(Str,4)), FormMain.PnjThumb((Parser(Str,8)='O'), round(CVal(Parser(Str,9)))), FormMain, FormMain.PanelMap, FormMain.ImageMapDragOver, FormMain.ImageThumbMouseMove);
        Pnj.SetType((Parser(Str,8)='O'), round(CVal(Parser(Str,9))), nil);
        Pnj.ObjID := round(CVal(Parser(Str,2)));

      end else if ( TypeObj = 'TPlayer' ) then begin
      // PLAYER
      // Obj;TPlayer;1;NewPlayer;Comments|;;0;0;
        Player := TPlayer.Create(Node, Parser(Str,3), ParseComment(Parser(Str,4)), FormMain.TeamThumb(round(CVal(Parser(Str,8)))), FormMain, FormMain.PanelMap, FormMain.ImageMapDragOver, FormMain.ImageThumbMouseMove);
        Player.SetTeam(round(CVal(Parser(Str,8))), nil);
        Player.ObjID := round(CVal(Parser(Str,2)));

      end else if ( TypeObj = 'TEvent' ) then begin
      // EVENT
        Event := TEvent.Create(Node, Parser(Str,3), ParseComment(Parser(Str,4)), Parser(Str,6), (Parser(Str,7) = 'O'));
        Event.TimePerc := CVal(Parser(Str,5));
        Event.ObjID := round(CVal(Parser(Str,2)));
      end;
    end;
  end;

  // Deuxime passe : liaison des objets
  for Boucle:=0 to Datas.count-1 do begin
    // lire une ligne
    Str := Datas[Boucle];
    // la decoder
    if ( Parser(Str,0) = 'Obj' ) then begin

      TypeObj := Parser(Str,1);
      // Tous les objets movables
      if (( TypeObj = 'TPnj' ) or
          ( TypeObj = 'TAera' ) or
          ( TypeObj = 'TPlayer' ) ) then begin
        MoveMovable(Str);
      end

    end;
  end;
  AllObjects.AlphaSort(True);
  FormMain.DrawTimeMap;
  if (assigned(TempNode)) then TempNode.Delete;
end;

// ----- CLIENT ------

procedure TDMod.OpenClient(IP: String; Port: Integer);
begin
  TempBuffer := '';
  Buffer.Clear;
  Client.Host := IP;
  Client.Port := Port;
  Client.Open;
end;

procedure TDMod.CloseClient;
begin
  if (Client.Active) then begin
    ToServer('!CHAT' + LF + FormMain.Nick.Text + ' disconnected...' + LF + '!END-CHAT' + LF);
    Client.Close;
  end;
end;

procedure TDMod.ClientConnect(Sender: TObject; Socket: TCustomWinSocket);
begin
  if (not Server.Active) then FormMain.CtrlTree.Items[CCTRL_CLI].ImageIndex := CICON_GREEN;
  ToServer('!CONNECT' + LF + FormMain.Nick.Text + LF + '!END-CONNECT' + LF);
end;
procedure TDMod.ClientDisconnect(Sender: TObject; Socket: TCustomWinSocket);
begin
  FormMain.ChatMemo.Lines.Append('Server disconnected...');
  FormMain.CtrlTree.Items[CCTRL_CLI].ImageIndex := CICON_RED;
end;
procedure TDMod.ClientError(Sender: TObject; Socket: TCustomWinSocket; ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
  FormMain.ChatMemo.Lines.Append('Server lost...');
  FormMain.CtrlTree.Items[CCTRL_CLI].ImageIndex := CICON_RED;
end;

procedure TDMod.ClientRead(Sender: TObject; Socket: TCustomWinSocket);
begin
  TempBuffer := TempBuffer + Socket.ReceiveText;
  Hash(TempBuffer, Buffer, LF);
  DoBufferCmd(Buffer);
end;

procedure TDmod.ToServer(Str: String);
begin
  if (Client.Active) then Client.Socket.SendText(Str);
end;


// ------- SERVEUR ---------

procedure TDmod.OpenServer(Port: Integer);
begin
  DMod.SaveDialog.DefaultExt := '*.evo';
  DMod.SaveDialog.FilterIndex := 2;
  SaveDialog.InitialDir := ExtractFilePath(Application.ExeName) + 'Saves';
  SaveDialog.Execute;
  if (DMod.SaveDialog.FileName <> '') then begin
    assignfile(ServerFile , SaveDialog.FileName);
    rewrite(ServerFile);

    TempBuffer := '';
    Buffer.Clear;
    NbConnected := 0;
    TimeCnt := 0;
    Server.Port := Port;
    Server.Open;
    OpenClient('127.0.0.1', Port);
  end;
end;

procedure TDmod.CloseServer;
begin
  if (Server.Active) then begin
    Client.Close;
    Server.Close;
    Try
      CloseFile(ServerFile);
    Except
    End;
  end;
  FormMain.CtrlTree.Items[CCTRL_SER].ImageIndex := CICON_RED;
  FormMain.CtrlTree.Items[CCTRL_CLI].ImageIndex := CICON_RED;
end;

procedure TDMod.ServerListen(Sender: TObject; Socket: TCustomWinSocket);
begin
  FormMain.CtrlTree.Items[CCTRL_SER].ImageIndex := CICON_GREEN;
end;
procedure TDMod.ServerClientConnect(Sender: TObject; Socket: TCustomWinSocket);
begin
  inc(NbConnected);
  if (NbConnected > 1) then FormMain.CtrlTree.Items[CCTRL_CLI].ImageIndex := CICON_GREEN;
end;
procedure TDMod.ServerClientDisconnect(Sender: TObject; Socket: TCustomWinSocket);
begin
  dec(NbConnected);
  if (NbConnected < 2) then FormMain.CtrlTree.Items[CCTRL_CLI].ImageIndex := CICON_RED;
end;

procedure TDMod.ServerAccept(Sender: TObject; Socket: TCustomWinSocket);
var Boucle: Integer;
    JDRObject: TJDRObject;
    Str: string;
begin
  Str := '!LOAD' + LF + FormMain.SaveClocksSerial + LF;
  for boucle := 0 to AllObjects.Count - 1 do begin
    JDRObject := AllObjects[boucle].Data;
    if (assigned(JDRObject)) then Str := Str + JDRObject.Serialization + LF;
  end;
  Str := Str + '!END-LOAD' + LF;
  if (FormMain.TimePlayBtn.Down) then Str := Str + '!TIMEDOWN' + LF else Str := Str + '!TIMEUP' + LF;
  if (Socket.RemoteAddress <> Socket.LocalAddress) then
    Socket.SendText(Str)
  else begin
    Write(ServerFile, Str);
    Flush(ServerFile);
  end;
end;

procedure TDMod.ServerClientRead(Sender: TObject; Socket: TCustomWinSocket);
var Boucle: Integer;
    Str: String;
begin
  Str := Socket.ReceiveText;
  Write(ServerFile, Str);
  Flush(ServerFile);
  for Boucle:=0 to Server.Socket.ActiveConnections-1 do begin
    if (Socket <> Server.Socket.Connections[Boucle]) then
      Server.Socket.Connections[Boucle].SendText(Str);
  end;
end;


procedure TDMod.TimerTimer(Sender: TObject);
var Boucle: Integer;
    Scan, VGameMinutes: Integer;
    JDRObject, JDRObject2: TJDRObject;
    Proxy: Integer;
    LowerProxy: Integer;
    Str: String;

    function ProxyScan(Obj1, Obj2: TPlayer): Integer;
    var Delta: Integer;
    begin
      Result := -1;
      if ((Obj1.Map <> nil) and (Obj1.Team <> Obj2.Team) and (Obj1.Map = Obj2.Map)) then begin
        Delta := round(sqrt(power(Abs(Obj1.PercX - Obj2.PercX), 2) + power(Abs(Obj1.PercY - Obj2.PercY), 2)) * 100);
        if (Delta <= Obj1.Map.ProxW) then Result := Delta;
      end;
    end;

begin
  // Gestion du temps
  if (((FormMain.TimePlayBtn.Down) or (TimeCnt = 0)) and (Server.Active)) then begin
    if ((TimeCnt = 240) or (TimeCnt = 0)) then begin
      FormMain.RealClock.Text := LeftStr(TimeToStr(GetTime), 5);
      VGameMinutes := Round(DMod.CVal(FormMain.GameMinutes.Text));
      if (TimeCnt <> 0) then begin
        inc(VGameMinutes);
        FormMain.GameMinutes.Text := DMod.CStr(VGameMinutes);
        FormMain.DrawTimeMap;
      end;
      ToServer('!CLOCK' + LF + FormMain.SaveClocksSerial + LF + '!END-CLOCK' + LF);
      TimeCnt := 1;
    end else inc(TimeCnt);
  end;

  FormMain.ProxyMemo.Clear;
  for Boucle:=0 to AllObjects.Count-1 do begin
    JDRObject := AllObjects[Boucle].Data;
    if (assigned(JDRObject)) then begin
      // Scan de proximit
      if ((JDRObject is TPlayer) and (Boucle < AllObjects.Count-1)) then begin
        LowerProxy := 110; // 110 est impossible car c'est un pourcentage
        for Scan:=Boucle+1 to AllObjects.Count-1 do begin
          JDRObject2 := AllObjects[Scan].Data;
          if (JDRObject2 is TPlayer) then begin
            Proxy := ProxyScan(TPlayer(JDRObject), TPlayer(JDRObject2));
            if (Proxy >= 0) then begin
              // Alerte de proximit
              if (Proxy < LowerProxy) then begin
                Str := JDRObject.Name + '-' + JDRObject2.Name + ' :' + TPlayer(JDRObject).Map.Name;
                LowerProxy := Proxy;
              end;
            end;
          end;
        end;
        if (LowerProxy <> 110) then FormMain.ProxyMemo.Items.Append(Str);
      end;
      // Scan de modif pour le rseau
      with JDRObject do begin
        if (Locked > 0) then begin
          dec(Locked);
          SetIcon;
        end;
        if (Modified) then begin
          ToServer('!MODIFY' + LF + Serialization + LF + '!END-MODIFY' + LF);
          Modified := False;
        end;
      end;
    end;
  end;
end;

end.

