unit UDevIL;

	// Benot Gerbier                                                                //
	// -> Fournit l'objet TDevilDLL encapsullant la DLL de devIL                     //
	// -> Fournit l'interface TDevilDLL_Image encapsullant une image gre par DevIL //

	// Usage:                                                                        //
	// 1- Crer une instance de TDevilDLL en lui passant les chemins de fichier des  //
	//    trois DLLs de DevIL (ceci charge les DLL).                                 //
	// 2- Utiliser sa fonction CreateImage pour crer une image vide, de type        //
	//    TDevilDLL_Image, dans DevIL.                                               //
	// 3- Les images se chargent depuis un fichier et s'assignent  un TBitmap avec, //
	//    respectivement, Load et AssignToBitmap.                                    //
	// 4- Lorsqu'une image n'a plus de rfrences, elle est automatiquement dcharge//
	//    de DevIL.                                                                  //
	// 5- Dtruire l'instance de TDevilDLL pour dcharger les DLL aprs usage.       //

interface

uses SysUtils, Graphics, Windows;

const

	// Messages des exceptions
	CDevilDLL_CanTLoad_DLL   : String = 'Given DLLs can''t be loaded.';
	CDevilDLL_CanTLoad_Proc  : String = 'Given DLLs doesn''t contain all the needed DevIL functions.';
	CDevilDLL_CanTCreateImage: String = 'Can''t create the new image.';
	CDevilDLL_CanTLoadImage  : String = 'Given image can''t be loaded.';
	CDevilDLL_CanTAssignImage: String = 'Can''t assign the image to the given bitmap.';
	CDevilDLL_CanTScaleImage : String = 'Can''t scale the image.';

	// Nom des fonctions dans la DLL
	CDevilDLL_InitIl        : PAnsiChar = 'ilInit';
	CDevilDLL_InitIlu       : PAnsiChar = 'iluInit';
	CDevilDLL_InitIlut      : PAnsiChar = 'ilutRenderer';
	CDevilDLL_ImageID_Create: PAnsiChar = 'ilGenImages';
	CDevilDLL_ImageID_Bind  : PAnsiChar = 'ilBindImage';
	CDevilDLL_ImageID_Delete: PAnsiChar = 'ilDeleteImages';
	CDevilDLL_ImageID_Copy  : PAnsiChar = 'ilCopyImage';
	CDevilDLL_LoadImage     : PAnsiChar = 'ilLoadImage';
	CDevilDLL_SaveImage     : PAnsiChar = 'ilSaveImage';
	CDevilDLL_GetImageInfo  : PAnsiChar = 'iluGetImageInfo';
	CDevilDLL_ImageParameter: PAnsiChar = 'iluImageParameter';
	CDevilDLL_Scale         : PAnsiChar = 'iluScale';
	CDevilDLL_ToBitmap      : PAnsiChar = 'ilutConvertToHBitmap';
	CDevilDLL_GetError      : PAnsiChar = 'ilGetError';

	// Constantes diverses DevIL
	CDevilDLL_NoError: Cardinal = 0; // IL_NO_ERROR
	CDevilDLL_Allegro: Cardinal = 2; // ILUT_ALLEGRO

	// Constantes diverses DevILU
  CDevilDLL_FILTER         : Cardinal = $2600;
  CDevilDLL_NEAREST				 : Cardinal = $2601;
  CDevilDLL_LINEAR				 : Cardinal = $2602;
  CDevilDLL_BILINEAR			 : Cardinal = $2603;
  CDevilDLL_SCALE_BOX			 : Cardinal = $2604;
  CDevilDLL_SCALE_TRIANGLE : Cardinal = $2605;
  CDevilDLL_SCALE_BELL		 : Cardinal = $2606;
  CDevilDLL_SCALE_BSPLINE	 : Cardinal = $2607;
  CDevilDLL_SCALE_LANCZOS3 : Cardinal = $2608;
  CDevilDLL_SCALE_MITCHELL : Cardinal = $2609;

type

	// Exceptions
	EDevilDLL_CanTLoad        = class( Exception );
	EDevilDLL_CanTCreateImage = class( Exception );
	EDevilDLL_CanTLoadImage   = class( Exception );
	EDevilDLL_CanTAssignImage = class( Exception );
	EDevilDLL_CanTScaleImage  = class( Exception );

  // Informations sur l'image
  TILinfo = record
    Id          : Cardinal;			// the image's id
    Data        : ^Byte; 	  		// the image's data
    Width       : Cardinal;			// the image's width
    Height      : Cardinal;			// the image's height
    Depth       : Cardinal;			// the image's depth
    Bpp         : Byte;			  	// bytes per pixel (not bits) of the image
    SizeOfData  : Cardinal;			// the total size of the data (in bytes)
    Format      : Cardinal;			// image format (in IL enum style)
    ILType      : Cardinal;			// image type (in IL enum style)
    Origin      : Cardinal;			// origin of the image
    Palette     : ^Byte;	    	// the image's palette
    PalType     : Cardinal;			// palette type
    PalSize     : Cardinal;			// palette size
    CubeFlags   : Cardinal;			// flags for what cube map sides are present
    NumNext     : Cardinal;			// number of images following
    NumMips     : Cardinal;			// number of mipmaps
    NumLayers   : Cardinal;			// number of layers
  end; PILinfo = ^TILinfo;

	// Types fonctionels utiliss pat TDevilDLL
	TDevilDLL_InitIL         = procedure;                                 stdcall;
	TDevilDLL_InitIlu        = procedure;                                 stdcall;
	TDevilDLL_InitIlut       = function ( Renderer: Cardinal ): Boolean;  stdcall;
	TDevilDLL_ImageID_Create = procedure( Count: Integer; IDs: Pointer ); stdcall;
	TDevilDLL_ImageID_Bind   = procedure( ID: Cardinal );                 stdcall;
	TDevilDLL_ImageID_Delete = procedure( Count: Integer; IDs: Pointer ); stdcall;
	TDevilDLL_ImageID_Copy   = procedure( ID: Cardinal );                 stdcall;
	TDevilDLL_LoadImage      = function ( FileName: PAnsiChar ): Boolean; stdcall;
	TDevilDLL_GetImageInfo   = procedure( ILinfo: PILinfo );              stdcall;
	TDevilDLL_ImageParameter = procedure( PName, Param: Integer );        stdcall;
	TDevilDLL_Scale          = function ( Width, Height, Depth: Cardinal ): Boolean; stdcall;
	TDevilDLL_ToBitmap       = function ( DC: HDC ): Cardinal;            stdcall;
	TDevilDLL_GetError       = function : Cardinal;                       stdcall;

	// Infos ncessaires  l'utilisation de la DLL DevIl
	TDevilDLL_Infos = record
		FileName_IL, FileName_ILU, FileName_ILUT: String;   // Nom des fichiers de DLL
		Handle_IL  , Handle_ILU  , Handle_ILUT  : Cardinal; // Handle des DLL

		Init_Il  : TDevilDLL_InitIL;   // Initialisation IL
		Init_Ilu : TDevilDLL_InitIlu;  // Initialisation ILU
		Init_Ilut: TDevilDLL_InitIlut; // Initialisation ILUT

		ImageID_Create: TDevilDLL_ImageID_Create; // Cre un/des identificateur(s) d'images
		ImageID_Bind  : TDevilDLL_ImageID_Bind;   // Lie l'image courante  un ID
		ImageID_Delete: TDevilDLL_ImageID_Delete; // Libre un ID et l'image associe
		ImageID_Copy  : TDevilDLL_ImageID_Copy  ; // Copy un ID dans l'image active

		LoadImage: TDevilDLL_LoadImage; // Charge une image en mmoire depuis un fichier

    GetImageInfo: TDevilDLL_GetImageInfo;     // Retourne des informations sur l'image
    ImageParameter: TDevilDLL_ImageParameter; // Configure le comportement de ILU
    Scale: TDevilDLL_Scale;                   // Redimensionne une image

		ToBitmap: TDevilDLL_ToBitmap; // Extrait l'image vers un device context

		GetError: TDevilDLL_GetError; // Retourne les erreurs DevIL
	end;

	// Une image stoque dans DevIL
	TDevilDLL_Image = interface
    function  ID: Cardinal;
		procedure Load          ( const FileName: String   );
    procedure Copy          ( Image: TDevilDLL_Image );
    function GetImageInfo : TILinfo;
    procedure Scale         ( const Filter, Width, Height, Depth: Cardinal );
		procedure AssignToBitmap( Target: Graphics.TBitmap );
    procedure Free;
	end;

	TDevilDLL = class
	private
		// Infos sur la DLL
		PInfos: TDevilDLL_Infos;
	public
		// Cration / Destruction
		constructor Create( const DLLName_IL, DLLName_ILU, DLLName_ILUT: String );
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;

		// Obtention d'un TDevilDLL_Image
		function CreateImage: TDevilDLL_Image;
	end;

implementation

type

	// Une image stoque dans DevIL
	PDevilDLL_Image = class( TInterfacedObject, TDevilDLL_Image )
	private
		// TDevilDLL parent
		PParent: TDevilDLL;
		PID: Cardinal;

	public
		// ID interne dans DevIL

		// Cration / Destruction
		constructor Create( Parent: TDevilDLL );
		procedure AfterConstruction; override;
		procedure BeforeDestruction; override;

		// Activation de l'image dans DevIL
		procedure Activate;

		// Fonctions de manipulation
    function  ID: Cardinal;
		procedure Load          ( const FileName: String   );
    procedure Copy          ( Image: TDevilDLL_Image );
    function GetImageInfo : TILinfo;
    procedure Scale         ( const Filter, Width, Height, Depth: Cardinal );
		procedure AssignToBitmap( Target: Graphics.TBitmap );
    procedure Free;
	end;


	// Cration / Destruction
	constructor TDevilDLL.Create( const DLLName_IL, DLLName_ILU, DLLName_ILUT: String );
	begin
		PInfos.FileName_IL   := DLLName_IL;
		PInfos.FileName_ILU  := DLLName_ILU;
		PInfos.FileName_ILUT := DLLName_ILUT;
		PInfos.Handle_IL     := 0;
		PInfos.Handle_ILU    := 0;
		PInfos.Handle_ILUT   := 0;
		inherited Create;
	end;


	// ------------------------------------------------------------
	// - TDevilDLL                                                -
	// ------------------------------------------------------------

	procedure TDevilDLL.AfterConstruction;
	begin
		// Chargement de la DLL et initialisation de PInfos
		PInfos.Handle_IL      := LoadLibrary( PAnsiChar(PInfos.FileName_IL  ) );                 if ( PInfos.Handle_IL   = 0                    ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_DLL) );
		PInfos.Handle_ILU     := LoadLibrary( PAnsiChar(PInfos.FileName_ILU ) );                 if ( PInfos.Handle_ILU  = 0                    ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_DLL) );
		PInfos.Handle_ILUT    := LoadLibrary( PAnsiChar(PInfos.FileName_ILUT) );                 if ( PInfos.Handle_ILUT = 0                    ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_DLL) );
		PInfos.Init_Il        := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_InitIl         ); if ( Assigned( PInfos.Init_Il        ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.Init_Ilu       := GetProcAddress( PInfos.Handle_ILU,  CDevilDLL_InitIlu        ); if ( Assigned( PInfos.Init_Ilu       ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.Init_Ilut      := GetProcAddress( PInfos.Handle_ILUT, CDevilDLL_InitIlut       ); if ( Assigned( PInfos.Init_Ilut      ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.ImageID_Create := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_ImageID_Create ); if ( Assigned( PInfos.ImageID_Create ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.ImageID_Bind   := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_ImageID_Bind   ); if ( Assigned( PInfos.ImageID_Bind   ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.ImageID_Delete := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_ImageID_Delete ); if ( Assigned( PInfos.ImageID_Delete ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.ImageID_Copy   := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_ImageID_Copy   ); if ( Assigned( PInfos.ImageID_Copy   ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.LoadImage      := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_LoadImage      ); if ( Assigned( PInfos.LoadImage      ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
    PInfos.GetImageInfo   := GetProcAddress( PInfos.Handle_ILU,  CDevilDLL_GetImageInfo   ); if ( Assigned( PInfos.GetImageInfo   ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.ImageParameter := GetProcAddress( PInfos.Handle_ILU,  CDevilDLL_ImageParameter ); if ( Assigned( PInfos.ImageParameter ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.Scale          := GetProcAddress( PInfos.Handle_ILU,  CDevilDLL_Scale          ); if ( Assigned( PInfos.Scale          ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.LoadImage      := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_LoadImage      ); if ( Assigned( PInfos.LoadImage      ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.ToBitmap       := GetProcAddress( PInfos.Handle_ILUT, CDevilDLL_ToBitmap       ); if ( Assigned( PInfos.ToBitmap       ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );
		PInfos.GetError       := GetProcAddress( PInfos.Handle_IL,   CDevilDLL_GetError       ); if ( Assigned( PInfos.GetError       ) = False ) then Raise( EDevilDLL_CanTLoad.Create(CDevilDLL_CanTLoad_Proc) );

		// Initialisation de DevIL
		PInfos.Init_Il;
		PInfos.Init_Ilu;
		PInfos.Init_Ilut( CDevilDLL_Allegro );

		inherited AfterConstruction;
	end;

	procedure TDevilDLL.BeforeDestruction;
	begin
		try
			// Dchargement de la DLL
			if ( PInfos.Handle_IL   <> 0 ) then FreeLibrary( PInfos.Handle_IL   );
			if ( PInfos.Handle_ILU  <> 0 ) then FreeLibrary( PInfos.Handle_ILU  );
			if ( PInfos.Handle_ILUT <> 0 ) then FreeLibrary( PInfos.Handle_ILUT );
		finally
			inherited BeforeDestruction;
		end;
	end;

	// Obtention d'un TDevilDLL_Image
	function TDevilDLL.CreateImage: TDevilDLL_Image;
	begin
		Result := PDevilDLL_Image.Create( Self );
	end;


	// ------------------------------------------------------------
	// - PDevilDLL_Image                                          -
	// ------------------------------------------------------------

	// Cration / Destruction
	constructor PDevilDLL_Image.Create( Parent: TDevilDLL );
	begin
		PParent := Parent;
		inherited Create;
	end;

	procedure PDevilDLL_Image.AfterConstruction;
	begin
		// Cration et liage de l'image
		PID := 0;
		PParent.PInfos.ImageID_Create( 1, @PID );
		PParent.PInfos.ImageID_Bind  ( PID     );

		// Vrification d'ventuelles erreurs
		if ( PParent.PInfos.GetError <> CDevilDLL_NoError ) then begin
			// On vide le buffer d'erreurs de DevIL...
			while( PParent.PInfos.GetError <> CDevilDLL_NoError ) do begin  end;

			// ...puis erreur
			Raise( EDevilDLL_CanTCreateImage.Create(CDevilDLL_CanTCreateImage) );
		end;
	end;

  function  PDevilDLL_Image.ID: Cardinal;
  begin
    Result := PID;
  end;

	procedure PDevilDLL_Image.BeforeDestruction;
	begin
		// Libration ventuelle de l'image
		if ( PID <> 0 ) then PParent.PInfos.ImageID_Delete( 1, @PID );

		// Vidange du buffer d'erreurs de DevIL
		while( PParent.PInfos.GetError <> CDevilDLL_NoError ) do begin  end;
	end;

	// Activation de l'image dans DevIL
	procedure PDevilDLL_Image.Activate;
	begin
		PParent.PInfos.ImageID_Bind( PID );
	end;

  procedure PDevilDLL_Image.Copy( Image: TDevilDLL_Image );
  begin
    Activate;
    PParent.PInfos.ImageID_Copy( Image.ID );
  end;

	// Fonctions de manipulation
	procedure PDevilDLL_Image.Load( const FileName: String  );
	begin
		// Activation de l'image
		Activate;

		// Chargement de l'image
		PParent.PInfos.LoadImage( PAnsiChar(FileName) );

		// Vrification d'ventuelles erreurs
		if ( PParent.PInfos.GetError <> CDevilDLL_NoError ) then begin
			// On vide le buffer d'erreurs de DevIL...
			while( PParent.PInfos.GetError <> CDevilDLL_NoError ) do begin  end;

			// ...puis erreur
			Raise( EDevilDLL_CanTLoadImage.Create(CDevilDLL_CanTLoadImage) );
		end;
	end;

	procedure PDevilDLL_Image.AssignToBitmap( Target: Graphics.TBitmap );
	begin
		// Erreur si la cible passe est Nil
		if ( Target = Nil ) then Raise( EDevilDLL_CanTAssignImage.Create(CDevilDLL_CanTAssignImage) );

		// Activation de l'image
		Activate;

    Target.FreeImage;
		// Assignation de l'image
		Target.Handle := PParent.PInfos.ToBitmap( Target.Canvas.Handle );

		// Vrification d'ventuelles erreurs
		if ( PParent.PInfos.GetError <> CDevilDLL_NoError ) then begin
			// On vide le buffer d'erreurs de DevIL...
			while( PParent.PInfos.GetError <> CDevilDLL_NoError ) do begin  end;

			// ...puis erreur
			Raise( EDevilDLL_CanTAssignImage.Create(CDevilDLL_CanTAssignImage) );
		end;
	end;

  procedure PDevilDLL_Image.Scale( const Filter, Width, Height, Depth: Cardinal );
  begin
    Activate;

    PParent.PInfos.ImageParameter( CDevilDLL_FILTER, Filter );
    PParent.PInfos.Scale( Width, Height, Depth );

		// Vrification d'ventuelles erreurs
		if ( PParent.PInfos.GetError <> CDevilDLL_NoError ) then begin
			// On vide le buffer d'erreurs de DevIL...
			while( PParent.PInfos.GetError <> CDevilDLL_NoError ) do begin  end;

			// ...puis erreur
			Raise( EDevilDLL_CanTScaleImage.Create(CDevilDLL_CanTScaleImage) );
		end;
  end;

  function PDevilDLL_Image.GetImageInfo : TILinfo;
  var ILinfo: TILinfo;
  begin
    Activate;

    PParent.PInfos.GetImageInfo( @ILinfo );
    Result := ILinfo;
  end;

  procedure PDevilDLL_Image.Free;
  begin
    Activate;

    PParent.PInfos.ImageID_Delete(1, @PID);
  end;

end.
