unit UJDRObjects;

interface

Uses Classes, Graphics, UDevil, StdCtrls, ExtCtrls, Controls, ComCtrls, SysUtils, Forms;

// MAPS

type
  TJDRObject = class
    private
    public
      AllObjects: TTreeNodes;
      SelfNode  : TTreeNode;
      Name    : string;
      Comments: TStringList;
      ObjID   : Integer;

      // Trigger reseau
      Locked: Integer;
      Modified: Boolean;

      procedure Init(VNode: TTreeNode; VName: string; VComments: TStrings);
      destructor Destroy; override;
      function Serialization: string; virtual;
      function LoadSerial(Datas: TStringList): Boolean; virtual;
      procedure SetIcon;
  end;

  // --- TDirectory : Objet rpertoires

  TDirectory = class(TJDRObject)
    public
      constructor Create(VNode: TTreeNode; VName: string);
      destructor Destroy; override;
      function Serialization: string; override;
      function LoadSerial(Datas: TStringList): Boolean; override;
  end;

  // --- TEvent : Objets vnements
  TEvent = class(TJDRObject)
    private
    public
      TimePerc: Double;
      Index: String;
      Active: Boolean;

      constructor Create(VNode: TTreeNode; VName: string; VComments: TStrings; VIndex: String; VActive: Boolean);
      destructor  Destroy; override;
      function Serialization: string; override;
      function LoadSerial(Datas: TStringList): Boolean; override;
  end;

  // --- TMap : objets cartes
  TMap = class(TJDRObject)
    private
      PProxW   : ShortInt;
      function  GetProxW: Integer;
      procedure SetProxW(Value: Integer);

    public
      Image   : string;
      Original: TDevilDLL_Image;
      OriginalInfo: TILinfo;
      Bitmap: TBitmap;
      SizeCoef: double;
      CurSizeX: Cardinal;
      CurSizeY: Cardinal;

      property ProxW : Integer read GetProxW write SetProxW;

      constructor Create(VNode: TTreeNode; VName: string; VComments: TStrings; Image: string; VProxW: Integer);
      procedure   LoadImage(VImage: string);
      procedure   DeclarePanel(Panel: TPanel);
      procedure   AffectSized(Size: Integer; Final: boolean);
      destructor  Destroy; override;

      function Serialization: string; override;
      function LoadSerial(Datas: TStringList): Boolean; override;
  end;

  // --- TPnj : objets Personnages non joueurs
  TMovable = class(TJDRObject)
    private
      ThumbParent: TWinControl;
      function  GetMapID: Integer;
      procedure SetMapID(Value: Integer);

    public
      MapThumb : TImage;
      MapLabel : TLabel;
      PercX    : Double;
      PercY    : Double;
      Map      : TMap;

      property MapID : Integer read GetMapID write SetMapID;

      constructor Create(VNode: TTreeNode; VName: string; VComments: TStrings; VImage: TImage; VOwner: TComponent; VParent: TWinControl; VDragOver: TDragOverEvent; VMouseMove: TMouseMoveEvent);
      procedure   SetPos(VMap: TMap; AX, AY: Integer);
      procedure   Display(Image: TImage);
      procedure   Hide;
      destructor  Destroy; override;

      function Serialization: string; override;
      function LoadSerial(Datas: TStringList): Boolean; override;
  end;

  TPnj = class(TMovable)
    public
      Monster : Boolean;
      Colour  : Integer;

      procedure SetType(VMonster: Boolean; VColour: Integer; VImage: TImage);
      function Serialization: string; override;
      function LoadSerial(Datas: TStringList): Boolean; override;
  end;

  TAera = class(TMovable)
  public
      function Serialization: string; override;
      function LoadSerial(Datas: TStringList): Boolean; override;
  end;

  TPlayer = class(TMovable)
  public
      Team: Integer;

      procedure SetTeam(Index: Integer; VImage: TImage);
      function Serialization: string; override;
      function LoadSerial(Datas: TStringList): Boolean; override;
  end;



implementation

uses UDmod, Main;

// --------------------
// ---  TJDRObject  ---
// --------------------

  procedure TJDRObject.Init(VNode: TTreeNode; VName: string; VComments: TStrings);
  var JDRObject: TJDRObject;
      Boucle: Integer;
      NewNode: TTreeNode;
      Ok: Boolean;
  begin
    // Charge les variables de l'objet
    Locked := 0;
    Modified := False;
    AllObjects := VNode.Owner;
    Name := VName;
    Comments := TStringList.Create;
    if (assigned(VComments)) then Comments.AddStrings(VComments);

    // Se cherche un identifiant unique
    repeat
      Ok := True;
      ObjID := Random(65534)+1;
      for Boucle:=0 to AllObjects.Count-1 do
        if (assigned(AllObjects[Boucle].Data)) then
          if (TJDRObject(AllObjects[Boucle].Data).ObjID = ObjID) then Ok := False;
    until (Ok);

    // S'ajoute dans l'arbre des objets
    JDRObject := VNode.Data;
    if ((not (Self is TDirectory)) and (JDRObject is TDirectory)) then
      NewNode := AllObjects.AddChildObject(VNode, Name, Self)
    else
      NewNode := AllObjects.AddObject(VNode, Name, Self);

    SelfNode := NewNode;
    SetIcon;
  end;

  destructor TJDRObject.Destroy;
  begin
    if assigned(SelfNode) then SelfNode.Delete;
    if assigned(Comments) then Comments.Free;
    inherited;
  end;

  procedure TJDRObject.SetIcon;
  var Icon: Integer;
  begin
    if (Locked > 0)              then Icon := CICON_LOCK
    else if (self is TDirectory) then Icon := CICON_FOLDER
    else if (self is TMap)       then Icon := CICON_MAP
    else if (self is TAera)      then Icon := CICON_AERA
    else if (self is TPnj)       then Icon := CICON_PNJ
    else if (self is TPlayer)    then Icon := CICON_PLAYER
    else if (self is TEvent)     then Icon := CICON_EVENT
    else Icon := -1;

    SelfNode.ImageIndex := Icon;
    SelfNode.SelectedIndex := Icon;
    SelfNode.TreeView.Refresh;
  end;

  function TJDRObject.Serialization: string;
  var boucle: Integer;
  begin
    result := DMod.CStr(ObjID) + ';' + DMod.BadChars(Name) + ';';
    for boucle:=0 to Comments.Count - 1 do result := result + DMod.BadChars(Comments[boucle]) + '|';
    Result := Result + ';';
  end;

  function TJDRObject.LoadSerial(Datas: TStringList): Boolean;
  begin
    if (Datas.Count > 2) then begin
      if (DMod.CVal(Datas[0]) = ObjID) then begin
        Datas.Delete(0);
        Name := Datas[0];
        SelfNode.Text := Name;
        Datas.Delete(0);
        Comments.Clear;
        Comments.AddStrings(DMod.ParseComment(Datas[0]));
        Datas.Delete(0);
        Result := True;
      end else Result := False;
    end else Result := False;
  end;

// ----------------------------------------
// ---  TDirectory : Objet rpertoires  ---
// ----------------------------------------

constructor TDirectory.Create(VNode: TTreeNode; VName: string);
begin
  Init(VNode, VName, nil);
end;

destructor  TDirectory.Destroy;
begin
  inherited;
end;

function  TDirectory.Serialization : String;
begin
  Result := 'Obj;TDirectory;' + inherited Serialization;
end;

function TDirectory.LoadSerial(Datas: TStringList): Boolean;
begin
  Result := False;
  if (Datas.Count = 5) then begin
    if ((Datas[0] = 'Obj') and (Datas[1] = 'TDirectory')) then begin
      Datas.Delete(0); Datas.Delete(0);
      if (Inherited LoadSerial(Datas)) then begin
        Result := True;
      end;
    end;
  end;
end;

// -----------------------------------
// ---  TEvent: objets venements  ---
// -----------------------------------

constructor TEvent.Create(VNode: TTreeNode; VName: string; VComments: TStrings; VIndex: String; VActive: Boolean);
begin
  Init(VNode, VName, VComments);
  TimePerc := -1;
  Index := VIndex;
  Active := VActive;
end;

destructor  TEvent.Destroy;
begin
  inherited;
end;

function TEvent.Serialization: string;
begin
  Result := 'Obj;TEvent;' + inherited Serialization + DMod.RStr(TimePerc) + ';' + DMod.BadChars(Index) + ';';
  if (Active) then Result := Result + 'O;' else Result := Result + 'N;';
end;

function TEVent.LoadSerial(Datas: TStringList): Boolean;
begin
  Result := False;
  if (Datas.Count > 7) then begin
    if ((Datas[0] = 'Obj') and (Datas[1] = 'TEvent')) then begin
      Datas.Delete(0); Datas.Delete(0);
      if (Inherited LoadSerial(Datas)) then begin
        TimePerc := DMod.CVal(Datas[0]);
        Datas.Delete(0);
        Index := Datas[0];
        Datas.Delete(0);
        Active := (Datas[0] = 'O');
        Datas.Delete(0);
        Result := True;
      end;
    end;
  end;
end;

// -----------------------------
// ---  TMap: objets cartes  ---
// -----------------------------

  constructor TMap.Create(VNode: TTreeNode; VName: string; VComments: TStrings; Image: string; VProxW: Integer);
  begin
    Init(VNode, VName, VComments);

    Original := DMod.Devil.CreateImage;
    Bitmap := TBitmap.Create;
    LoadImage(Image);
    ProxW := VProxW;
  end;

  procedure TMap.LoadImage(VImage: string);
  begin
    if (Image <> VImage) then begin
      Bitmap.FreeImage;
      Image := VImage;
      SizeCoef := 0;
      Original.Load(ExtractFilePath(Application.ExeName) + 'Maps\' + VImage);
      OriginalInfo := Original.GetImageInfo;
      CurSizeX := OriginalInfo.Width;
      CurSizeY := OriginalInfo.Height;
    end;
  end;

  procedure  TMap.DeclarePanel(Panel: TPanel);
  var DX, DY: double;
  begin
    DX := Panel.Width / OriginalInfo.Width;
    DY := Panel.Height / OriginalInfo.Height;
    if (DX > DY) then SizeCoef := DY else SizeCoef := DX;
    If (SizeCoef > 1) then SizeCoef := 1;
    Original.AssignToBitmap(Bitmap);
  end;

  procedure TMap.AffectSized(Size: Integer; Final: boolean);
  var Filter: Cardinal;
      CopyImg: TDevilDLL_Image;
      Coef: Double;
  begin
    Coef := 1;
//    Filter := CDevilDLL_BILINEAR;
    Filter := CDevilDLL_SCALE_LANCZOS3;
    if ( final ) then begin
      CopyImg := DMod.Devil.CreateImage;
      CopyImg.Copy(Original);
    end;

    if ( Size <> 0 ) then begin
      if ( Size > 0 ) then Coef := 1 + (( SizeCoef - 1 ) * (Size / 100))
                      else Coef := 1 + ( -Size / 100 );
      if (Coef <> 1) then begin
        CurSizeX := trunc(OriginalInfo.Width * Coef);
        CurSizeY := trunc(OriginalInfo.Height * Coef);
        if (final) then CopyImg.Scale(Filter, CurSizeX, CurSizeY, OriginalInfo.Depth );
      end;
    end;
    if ( Coef = 1 ) then begin
      CurSizeX := OriginalInfo.Width;
      CurSizeY := OriginalInfo.Height;
    end;
    if ( final ) then begin
      CopyImg.AssignToBitmap(Bitmap);
      CopyImg.Free;
    end;
  end;

  destructor TMap.Destroy;
  begin
    Original := nil;
    if assigned(Bitmap) then Bitmap.Free;
    inherited;
  end;

  function TMap.Serialization: string;
  begin
    Result := 'Obj;TMap;' + inherited Serialization + Image + ';' + DMod.CStr(ProxW) + ';';
  end;

  function TMap.LoadSerial(Datas: TStringList): Boolean;
  begin
    Result := False;
    if (Datas.Count > 3) then begin
      if ((Datas[0] = 'Obj') and (Datas[1] = 'TMap')) then begin
        Datas.Delete(0); Datas.Delete(0);
        if (Inherited LoadSerial(Datas)) then begin
          LoadImage(Datas[0]);
          Datas.Delete(0);
          ProxW := Round(DMod.CVal(Datas[0]));
          Datas.Delete(0);
          Result := True;
        end;
      end;
    end;
  end;

  // PROPERTY

  function  TMap.GetProxW: Integer;
  begin
    Result := PProxW;
  end;

  procedure TMap.SetProxW(Value: Integer);
  begin
    if (Value > 100) then Value := 100
    else if (Value < 0) then Value := 0;

    PProxW := Value;
  end;

// --------------------------------------
// ---  TMovable: objets dplacables  ---
// --------------------------------------

  constructor TMovable.Create(VNode: TTreeNode; VName: string; VComments: TStrings; VImage: TImage; VOwner: TComponent; VParent: TWinControl; VDragOver: TDragOverEvent; VMouseMove: TMouseMoveEvent);
  begin
    Init(VNode, VName, VComments);

    MapThumb := TImage.Create( VOwner );
    MapThumb.Enabled := False;
    MapThumb.Visible := False;
    MapThumb.DragMode := dmAutomatic;
    MapThumb.OnDragOver := VDragOver;
    MapThumb.OnMouseMove := VMouseMove;
    MapThumb.Tag := Integer(Self);
    MapThumb.Picture := VImage.Picture;
    MapThumb.Width := VImage.Width;
    MapThumb.Height := VImage.Height;
//    MapThumb.Parent := VParent;

    MapLabel := TLabel.create( VOwner );
    MapLabel.Transparent := True;
    MapLabel.Enabled := False;
    MapLabel.Visible := False;
    MapLabel.Tag := Integer(Self);
    MapLabel.Caption := Name;
//    MapLabel.Parent := VParent;

    ThumbParent := VParent;
    Map := nil;
    PercX := 0;
    PercY := 0;
  end;

  procedure TMovable.SetPos(VMap: TMap; AX, AY: Integer);
  var AvMap: TMap;
      AvPercX, AvPercY: Double;
  begin
    if (VMap <> nil) then begin
      AvMap := Map;
      AvPercX := PercX;
      AvPercY := PercY;

      Map := VMap;
      PercX := AX / VMap.CurSizeX;
      PercY := AY / VMap.CurSizeY;
      If (PercX < 0) then PercX := 0;
      If (PercX > 1) then PercX := 1;
      If (PercY < 0) then PercY := 0;
      If (PercY > 1) then PercY := 1;

      if ((AvMap <> Map) or (AvPercX <> PercX) or (AvPercY <> PercY)) then Modified := True;
    end;
  end;

  procedure TMovable.Display(Image: TImage);
  begin
    if (Map <> nil) then begin
      with MapThumb do begin
        Left := trunc(PercX * Map.CurSizeX) + Image.Left - (Width div 2);
        Top  := trunc(PercY * Map.CurSizeY) + Image.Top - (Height div 2);
        Enabled := True;
        Visible := True;
        Parent  := ThumbParent;
      end;
      with MapLabel do begin
        Left := MapThumb.Left + 16;
        Top  := MapThumb.Top;
        Visible := True;
        Parent  := ThumbParent;
      end;
    end;
  end;

  procedure TMovable.Hide;
  begin
    MapThumb.Parent := nil;
    MapThumb.Enabled := False;
    MapThumb.Visible := False;

    MapLabel.Parent := nil;
    MapLabel.Visible := False;
  end;

  destructor TMovable.Destroy;
  begin
    if assigned(MapThumb) then MapThumb.Free;
    if assigned(MapLabel) then MapLabel.Free;
    inherited;
  end;

  function  TMovable.Serialization : String;
  begin
    Result := inherited Serialization;
    if ( Map <> nil ) then begin
      Result := Result + DMod.CStr( Map.ObjID ) + ';';
      Result := Result + DMod.RStr( PercX ) + ';';
      Result := Result + DMod.RStr( PercY ) + ';';
    end else
      Result := Result + ';0;0;';
  end;

  function TMovable.LoadSerial(Datas: TStringList): Boolean;
  begin
    Result := False;
    if (Datas.Count > 5) then begin
      if (Inherited LoadSerial(Datas)) then begin
        MapID := round(DMod.CVal(Datas[0]));
        Datas.Delete(0);
        PercX := DMod.CVal(Datas[0]);
        Datas.Delete(0);
        PercY := DMod.CVal(Datas[0]);
        Datas.Delete(0);
        MapLabel.Caption := Name;
        Result := True;
      end;
    end;
  end;

  // PROPRIETES

  function TMovable.GetMapID: Integer;
  begin
    if (assigned(Map)) then
      Result := Map.ObjID
    else
      Result := 0;
  end;

  procedure TMovable.SetMapID(Value: Integer);
  var Obj: TObject;
  begin
    Obj := DMod.FindJDRObject(Value);
    if (Obj is TMap) then begin
      Map := TMap(Obj);
    end else Map := nil;
  end;



// -------------------------------
// ---  TPnj / TAera / TPlayer ---
// -------------------------------

  procedure TPnj.SetType(VMonster: Boolean; VColour: Integer; VImage: TImage);
  begin
    if ((VMonster <> Monster) or (VColour <> Colour)) then begin
      Monster := VMonster;
      Colour := VColour;
      if (assigned(VImage)) then MapThumb.Picture := VImage.Picture;
    end;
  end;

  function  TPnj.Serialization : String;
  begin
    Result := 'Obj;TPnj;' + inherited Serialization;
    if (Monster) then Result := Result + 'O;' else Result := Result + 'N;';
    Result := Result + DMod.CStr(Colour) + ';'
  end;

  function TPnj.LoadSerial(Datas: TStringList): Boolean;
  var Monster: Boolean;
      Colour : Integer;
  begin
    Result := False;
    if (Datas.Count = 10) then begin
      if ((Datas[0] = 'Obj') and (Datas[1] = 'TPnj')) then begin
        Datas.Delete(0); Datas.Delete(0);
        if (Inherited LoadSerial(Datas)) then Result := True;
        Monster := (Datas[0]='O');
        Colour := Round(Dmod.CVal(Datas[1]));
        SetType(Monster, Colour, FormMain.PnjThumb(Monster, Colour));
        Datas.Delete(0); Datas.Delete(0);
      end;
    end;
  end;

  function  TAera.Serialization : String;
  begin
    Result := 'Obj;TAera;' + inherited Serialization;
  end;

  function TAera.LoadSerial(Datas: TStringList): Boolean;
  begin
    Result := False;
    if (Datas.Count = 8) then begin
      if ((Datas[0] = 'Obj') and (Datas[1] = 'TAera')) then begin
        Datas.Delete(0); Datas.Delete(0);
        if (Inherited LoadSerial(Datas)) then Result := True;
      end;
    end;
  end;

  procedure TPlayer.SetTeam(Index: Integer; VImage: TImage);
  begin
    if (Team <> Index) then begin
      Team := Index;
      if (assigned(VImage)) then MapThumb.Picture := VImage.Picture;
    end;
  end;

  function  TPlayer.Serialization : String;
  begin
    Result := 'Obj;TPlayer;' + inherited Serialization;
    Result := Result + DMod.CStr(Team) + ';'
  end;

  function TPlayer.LoadSerial(Datas: TStringList): Boolean;
  var VTeam: Integer;
  begin
    Result := False;
    if (Datas.Count = 9) then begin
      if ((Datas[0] = 'Obj') and (Datas[1] = 'TPlayer')) then begin
        Datas.Delete(0); Datas.Delete(0);
        if (Inherited LoadSerial(Datas)) then begin
          VTeam := Round(Dmod.CVal(Datas[0]));
          SetTeam(VTeam, FormMain.TeamThumb(VTeam));
          Datas.Delete(0);
          Result := True;
        end;
      end;
    end;
  end;

end.

